:- module(aux, [
                bagof0/3, setof0/3,
		nodup/2,
                vars/2,
                var_occ/2,
                    term_rename/2,
                    get_term/2,
                    get_term/4,
		    match/3, ask_number/4,ask_oneof/3, yesno/1,read_sequence/3,
                    writel/1
               ]).

:- use_module(ordsets,	[
	ord_add_element/3,
	ord_subset/2,
	ord_subtract/3,
	ord_union/3
   ]).

:- use_module(substitution,[sub_domain/2]).


/*term_rename(+A,-B) -B is a fresh copy of +A.*/

term_rename(A,B) :-
    asserta('$'(A)),
    retract('$'(B)).

setof0(X,Goal,SX) :- setof(X,Goal,SX1) -> SX=SX1; SX=[].
bagof0(X,Goal,SX) :- bagof(X,Goal,SX1) -> SX=SX1; SX=[].

/*get_term(-Term,+Hint) prints Hint and reads Term*/

get_term(Term,Hint) :- 
        writel(Hint),
        ( read(Term),!
         ;nl,get_term(Term,Hint)
        ),
        nl.

writel([]).
writel([H|L]) :-
        ( H==nl -> nl ; write(H)),
        writel(L).

%get_term/4

get_term(Stream,Hint,Term,SYMBTB) :-
    writel(Hint),
    read_term(Stream,[syntax_errors(dec10),variable_names(NL)],Term),
    vars(Term,VL),
    nameListToSub(VL,NL,SYMBTB).


/*nameListToSub(+VL,+NL,-SUB) -SUB  is  a substitution over  domain VL
with variables assigend  their names as defined  in NL or an anonymous
names if undefined in NL.*/

nameListToSub(VL,NL,Sub) :-
    nameListToSub_1(NL,Sub1),
    sub_domain(Sub1,VL1),
    ord_subtract(VL,VL1,VL2),
    anonymous_sub(VL2,Sub2),
    ord_union(Sub1,Sub2,Sub). 

nameListToSub_1([],[]).
nameListToSub_1([Name=Var|NL1],Sub) :-
    nameListToSub_1(NL1,Sub1),
    ord_add_element(Sub1,Var/Name,Sub).

anonymous_sub([],[]).
anonymous_sub([Var|Vars],[Var/'_'|Sub]) :-
    anonymous_sub(Vars,Sub).

/*vars(+TERM,-VARS) VARS is the set of variables in TERM and is
  ordered*/

vars(T,Bag) :-
    vars(T,[],Bag).

vars(X,Bag,Bag) :-	
    atomic(X),
    !.
vars(X,Bag,NewBag) :-
    var(X),
    !,
    (   
        \+ ord_subset([X],Bag)
    -> 
        ord_add_element(Bag,X,NewBag)
    ;  
        NewBag = Bag
    ).
vars(X,Bag,NewBag) :-	
%    \+ atomic(X),
%    \+ var(X),
    !,
    X=..[_|Args],
    vars_list(Args,Bag,NewBag).

vars_list([],Bag,Bag).
vars_list([Arg1|Args2],Bag,NewBag) :-	
    vars_list(Args2,Bag,NewBag1),
    vars(Arg1,NewBag1,NewBag).

%%%%%%%%%%%

/*var_occ(+TERM,-VAROCC) VAROCC  is  an  ordered list  of pairs  of  a
variable  and its number of occurrences in TERM. Each variable in TERM
corresponds to a pair in VAROCC*/

var_occ(T,Bag) :-
    var_occ(T,[],Bag).

var_occ(X,Bag,Bag) :-	
    atomic(X),
    !.
var_occ(X,Bag,NewBag) :-
    var(X),
    !,
    insert(Bag,X,NewBag).
var_occ(X,Bag,NewBag) :-	
%    \+ atomic(X),
%    \+ var(X),
    !,
    X=..[_|Args],
    var_occ_list(Args,Bag,NewBag).

var_occ_list([],Bag,Bag).
var_occ_list([Arg1|Args2],Bag,NewBag) :-	
    var_occ_list(Args2,Bag,NewBag1),
    var_occ(Arg1,NewBag1,NewBag).

insert([],Var,[(Var,1)]).
insert([(Var1,Occ1)|Bag],Var,[(Var1,NewOcc1)|Bag]) :-
    	Var == Var1,
	!,
	NewOcc1 is Occ1+1.
insert([(Var1,Occ1)|Bag],Var,[(Var,1),(Var1,Occ1)|Bag]) :-
    	Var @< Var1,
	!.
insert([(Var1,Occ1)|Bag],Var,[(Var1,Occ1)|NewBag]) :-
    	%Var >@ Var1,
	!,
	insert(Bag,Var,NewBag).


/* match(+Patterns, +Prefix, -Matches) Matches is the subset of Patterns
   which have a prefix Prefix. */


match(Constants, Chars, Matches) :-
        match(Constants, Chars, [], Matches).

match([], _, Matches, Matches).
match([Constant|Constants], Chars, Matches0, Matches) :-
	(
	        \+ (name(Constant,L), \+ append(Chars,_,L))
        ->
		match(Constants,Chars,[Constant|Matches0],Matches)
	;
	        match(Constants, Chars, Matches0, Matches)
        ).


/* read_sequence(TSet,-Chars,-T) read a sequence -Chars of characters which is
   terminated by a terminator -T in TSet. */

read_sequence(TSet, Chars, T) :-
    discard_space, 
    read_until(TSet,Chars, T).

discard_space :- 
    peek_byte(C0),( C0=32 -> get0(_),discard_space; true ).

read_until(TSet,Chs,T) :- 
    get0(Ch),
    ( 
        member(Ch,TSet) 
    -> 
        Chs=[], T=Ch
    ; 
        read_until(TSet,Chs1,T), 
        Chs=[Ch|Chs1]
    ). 

/* ask_oneof(+Prompt,+Choices,-Choice) */

ask_oneof(Prompt,Choices,Choice) :-
    write(Prompt),
    read_sequence([32,46,10],Input,_Terminator),
    match(Choices,Input,Match),
    (
         Match = [Choice]
     -> 
         true
     ; 
         nl,
         ask_oneof(Prompt,Choices,Choice)
    ).


/* yesno(+Sentence) */

yesno(Sentence) :-
    writel(Sentence),write('? '),
    read_sequence([32,46,10],Input,_Terminator),
    match([yes,no],Input,Match),
    (
         Match = []
     ->
         nl, 
         yesno(Sentence)
     ;
         Match = [yes]
     -> 
         true
    ).

/* ask_number(+Prompt,+Min,+Max,-Num) */

ask_number(Prompt,Min,Max,Num) :-
    write(Prompt), write('? '),
    read_sequence([32,46,10],Input,_Terminator),
    ( 
         name(Num,Input), integer(Num), Min =< Num, Num =< Max
    -> 
         true
     ;
         nl,
         ask_number(Prompt,Min,Max,Num)
    ).

/* nodup(+Si,-So) So is the list containing all members of Si without
   duplication*/

nodup([],[]).
nodup([X|Si],So) :-
    nodup(Si,So_),
    ( in(X,So_) -> So=So_ ; So=[X|So_] ).

in(X,[Y|L]) :- 
    ( X==Y -> true; in(X,L) ).
