:- module(diagnoser,[diagnoser/3]). 

:- use_module(trace,[
       icppt/3,
       icppt_transformer/3,
       icppt_RootImplication/2,
       icpst/3,
       icpst_divisor/5,
       icpst_transformer1/2,       
       icpst_transformer2/2,
       aet/3]).

:- use_module(store,[pg_AddCLS/3,cls_CONS/4]).

:- use_module(meta,[solve_db/3]).

:- use_module(goal,[goal_select/3]).

:- use_module(oracle,[admissible/3,
                      examine_solutions/4,
                      satisfiable/2,
                      valid/2]).


:- use_module(strategies,[non_root_node/3,branch_node/3]).

:- use_module(substitution,[sub_apply/3,sub_genSymbTb/3,sub_restrict/3,variant/2]).

:- use_module(orderedTree,[
        ordTree_display/2,
        ordTree_getSubTree/3,
        ordTree_removeSubTree/3,
        ordTree_shrinkSubTree/3,
        ordTree_rootNode/2,
        ordTree_height/2,
        ordTree_weight/2,
	ordTree_leftmost/2,
	ordTree_rightmost/2
              ]).

:- use_module(meta,[solve_db/3]).

:- use_module(parameters,[read_parameter_value/2]).

:- use_module(aux,[term_rename/2,vars/2,writel/1]).

/*diagnoser(+P,+S,-B) +P is a program,+S is  a  symptom of +P, -B is a bug.*/

diagnoser(P,declarative(miss,A),B) :-
    (
        read_parameter_value(query_instance,yes) 
    ->
        insufficiency(P,A,B)
    ; 
        incompleteness(P,A,B)
    ).
diagnoser(P,declarative(wrong,A),B) :-
    inconsistency(P,A,B).
diagnoser(P,operational(Cat,A),B) :-
    operational(P,Cat,A,B).    

/*

inconsistency(+P,+A,-D)  +P is  a program,  +A an atom  and D either a  clause
instance or an atom.  By the soundness and the completeness theorems, if +A is
an  inconsistency symptom  then inconsistency(P,A,D)  succeeds  with  D  being
either an inconsistent clause instance or an incompletetly covered atom.

*/

inconsistency(P,A,D) :-
    writel([A, ' is an inconsistency symptom, diagnosing...',nl]),
    icppt(P,A,T1), % ordTree_display(T1,0),
    icppt_transformer(T1,P,T), % ordTree_display(T,0),
    !,
    invalid_impl(T,P,C),
    (
        C=(\+A1:-true) 
    -> 
        diagnoser(P,declarative(miss,A1),D)
    ; 
        D=inconsistent(C)
        %,
        %nl,write('Report: inconsistent clause instance:'),
        %nl,tab(5),write(C),nl
    ).

/*

invalid_impl(+T,+P,-C)   +T  is  a  literal-labelled   tree  and  -C  root
impl of a subtree of T. The root impl of a literal-labelled tree
is the clause whose head is the root node and whose premises are the son nodes
of the root  node.   By  the  soundness and the completeness lemmas, if T is a
literal-labelled tree such  that its root is invalid  and its leaves are valid
then  invalid(impl(T,C) succeeds with C being the root impl of a
subtree and C being invalid.

*/


invalid_impl(T,_P,C) :-
    ordTree_height(T,1),
    !,
    icppt_RootImplication(T,C).
invalid_impl(T,P,C) :-
%   \+ ordTree_height(T,1),
    branch_node(T,node(Id,L,_,_)),
    !,
    (    
        valid(L,P) 
    -> 
        ordTree_shrinkSubTree(T,Id,Tnew)
    ; 
        ordTree_getSubTree(T,Id,Tnew)
    ),
    invalid_impl(Tnew,P,C).

/*

insufficiency(+P,+A,-D) +P is a program, +A an atom and  -D is either a clause
instance or an atom. By the soundness and the the completeness theorems, if +A
is an insufficiency symptom  then insufficiency(P,A,D) succeeds  with -D being
either an inconsistent clause instance or an incompletely covered atom.

*/

insufficiency(P,A,D) :-
    writel([A, ' is an insufficiency symptom, diagnosing...',nl]),
    icpst(P,A,T1), % ordTree_display(T1,0),
    icpst_transformer1(T1,T2), % ordTree_display(T2,0),
    icpst_transformer2(T2,T),  % ordTree_display(T,0),
    critical(T,P,node(_,W/_,_,_)),
    goal_select(W,L,_),
    (    
        L=(\+A1) 
    -> 
        diagnoser(P,declarative(wrong,A1),D)
    ; 
        D=uncovered(L)
%        ,nl,write('Report: incompletely covered atom:'),
%        nl,tab(5),write(L),nl
    ). 

/* 

critical(+T,+P,-N) +T is a goal-labelled tree and -N is a node of +T.
By the soundness and the completeness lemmas, if T is a goal labelled
tree with root '<-'(W) such that W is satisfiable and none of the
leaves of +T is an empty goal, then critical succeeds with -N being a
critical node.  A node '<-(W') of T is a critical node if W' is
satisfiable and, for each son '<-'(W'') of '<-(W'), W'' is
unsatisfiable.

*/

critical(T,_P,N) :- 
    ordTree_height(T,0),
    !,
    ordTree_rootNode(T,N).
critical(T,P,N) :-
%   \+ ordTree_height(T,0),
    non_root_node(T,node(Id,W/_,_,_)),
    term_rename(W,W0),
    !,
    (   
        satisfiable(W0,P) 
    -> 
        ordTree_getSubTree(T,Id,Tnew)
    ;  
        ordTree_removeSubTree(T,Id,Tnew)
    ),
    critical(Tnew,P,N).

incompleteness(P,A,D) :-
    writel([A, ' is an incompleteness symptom, diagnosing...',nl]),
    icpst(P,A,T1), % ordTree_display(T1,0),
    icpst_transformer1(T1,T2), % ordTree_display(T2,0),
    icpst_transformer2(T2,T), % ordTree_display(T,0),
    critical_1(T,P,N),
    (
        N=node(_,W/_,_,_)
    ->
        goal_select(W,L,_),
        (    
            L=(\+A1) 
        -> 
            inconsistency(P,A1,D)
        ; 
            D=uncovered(L)
        )
    ;   
        N = declarative(wrong,A1)
    ->
        inconsistency(P,A1,D)
    ;
        N = declarative(miss,A1)
    ->
        incompleteness(P,A1,D)
    ). 

critical_1(T,P,N) :-   % ordTree_display(T,0),nl,nl,
    (
        non_root_node(T,node(Id,W/_,_,_))
    ->
        goal_select(W,L,_),
        icpst_divisor(T,Id,CPS1,Sa,CPS2), 
        examine_solutions(L,Sa,P,R), 
        (    
            R = complete
        -> 
            critical_1(CPS2,P,N)
        ;  
            R = incomplete
        ->
            critical_1(CPS1,P,N)
        ; 
            N = R % R=declarative(miss,A) or R=declarative(wrong,A)
        )
    ;
        ordTree_rootNode(T,N)
    ). 


/* choose right search strategies */

non_root_node(T,N) :- 
    read_parameter_value(miss_search_strategy,STRATEGY),
    non_root_node(STRATEGY,T,N).
branch_node(T,N) :-
    read_parameter_value(wrong_search_strategy,STRATEGY),
    branch_node(STRATEGY,T,N).


operational(P,_,A,D) :-
    (  
        aet(P,A,t(Id,Lb,Ht,Wt,CHN))
    -> 
	ordTree_display(t(Id,Lb,Ht,Wt,CHN),0),
        (
	    segment(t(Id,Lb,Ht,Wt,CHN),P,S)
    	->
            diverge(S,P,D)
        ;
            D = unknown

        )
    ; 
       D = unknown,
       writel(['No detected ',operational,' error.',nl])
    ).


segment(T,P,S)            :-
    \+ ordTree_height(T,0),
    ordTree_rightmost(T,node(Id,(Pt,Sub),_,_)),
    (   
        admissible(P,Pt,Sub)
    ->  
        ordTree_getSubTree(T,Id,T1),
        segment(T1,P,S)
    ;   
        ordTree_shrinkSubTree(T,Id,S)
    ).

diverge(S,_P,diverging_invocation(PS0,PS1)) :-
    ordTree_weight(S,2),
    ordTree_rootNode(S,node(_,PS0,_,_)),
    ordTree_rightmost(S,node(_,PS1,_,_)).
diverge(S,P,D)            :-
    \+ ordTree_weight(S,2),
    ordTree_leftmost(S,node(Id,L,_,_)),
    (
        valid(L,P)
    ->
        ordTree_removeSubTree(S,Id,S1),
        diverge(S1,P,D)
    ;
        (
	    L = \+(A)
	->
	    insufficiency(P,A,D)
	;
	    inconsistency(P,L,D)
	)
    ).
