:- module(orderedTree,[
            ordTree_display/2,
            ordTree_constructor/4,
            ordTree_getSubTree/3,
            ordTree_removeSubTree/3,
            ordTree_shrinkSubTree/3,
            ordTree_rootNode/2,
            ordTree_getNode/2,
            ordTree_getLeafNode/2,
            ordTree_getChildOfRoot/2,
            ordTree_weight/2,
            ordTree_height/2,
                node_isLeaf/1,
                node_isSolutionNode/1,
            forest_containingTree/5, 
            forest_weight/2,
            forest_height/2,
	    ordTree_leftmost/2,
	    ordTree_rightmost/2
         ]).


/*
SPECIFICATION
-------------

This module contains procedures for basic operations on tree.  We
number every node of a tree in such a way that if node X is traversed
in pre-order before node Y then the number assigned to X is less than
that assigned to Y. We also associated each node of a tree with two
other numbers the height of the subtree rooted at the node and the
weight of the subtree rooted at the node. The relationship beteew a
node and its children is encapsulated in the structure representing
the subtree rooted at that node.  A tree is represented as
t(Id,L,H,W,[t(....),t(....),...])  where Id is the number assigned to
the root of the tree, L the label associated with the root, H the
height of the tree, W the wight of the tree, t(....) in the list is a
subtree of the tree in that order and the root of the subtree is a
child of the root of the tree. An empty tree is represented as nil.
For instance, structure t(0,x,3,6, [t(1,y,2,4, [t(2,z,1,2,
[t(3,a,0,1,[])]),t(4,b,0,1,[])]), t(5,c,0,1,[])]) represents the
following tree

 x
 |--y
 |  |--z
 |  |  |--a
 |  |--b
 |--c
 
An isolated  node in a  tree  is represented  as  node(Id,L,H,W) where
Id,L,H and W have the same meaning as the above.

PUBLIC ordTree_display(+Tree,+Indentation)
       ordTree_getSubTree(+Tree,+Node_Id,-SubTree)
       ordTree_removeSubTree(+Tree,+Node_Id,-NewTree)
       ordTree_shrinkSubTree(+Tree,+Node_Id,-NewTree)
       forest_weight(+ListOfTrees,-SumOfWeights)
       forest_height(+ListOfTrees,-MaximumHeight)
       ordTree_weight(+Tree,-Weight)
       ordTree_height(+Tree,-Height)
*/


node_isLeaf(node(_,_,0,1)).
node_isSolutionNode(node(_,true/_,0,1)).


/* 

ordTree_display(+Tree,+SP) displays a tree +Tree with +SP being the
number of spaces indenting the root of +Tree.

*/

ordTree_display(nil,_).  
ordTree_display(t(Id,L,H,W,SubTrees),SP) :-
    write(Id),tab(2),tab(SP),
    write(L),write(' '),write(H),write(' '),write(W),nl,
    SP1 is SP+2,
    subtrees_display(SubTrees,SP1).

subtrees_display([],_).
subtrees_display([Tree|Forest],SP) :-
    ordTree_display(Tree,SP),
    subtrees_display(Forest,SP).        

/*

ordTree_constructor(+Id,+Label,+SubTrees,-Tree) The root of -Tree is
identified with +Id and labelled with +Label. The subtrees of the root
are +SubTrees, the weight of -Tree is one plus the sum of the weights
of the trees in +SubTrees, and the height of -Tree is one plus the
maximum if the heights of the trees in +SubTrees.

*/

ordTree_constructor(Id,Label,Subtrees,t(Id,Label,H,W,Subtrees)) :-
    forest_height(Subtrees,H1), H is H1+1,
    forest_weight(Subtrees,W1), W is W1+1.

/* 

ordTree_getSubTree(+Tree,+Node_Id,-SubTree) succeeds with -SubTree is the
subtree of +Tree rooted at the node with identifier +Node_Id iff
+Node_Id identifies a node of +Tree.

*/

ordTree_getSubTree(t(Id0,L0,W0,H0,CHN0),Id,SubTree) :-
    (  
        Id0==Id 
    -> 
        SubTree=t(Id0,L0,W0,H0,CHN0)
    ; 
        forest_containingTree(CHN0,Id,_,CH0,_),
        ordTree_getSubTree(CH0,Id,SubTree)
    ).

/* 

ordTree_removeSubTree(+Tree,+Node,-TreeNew) succeeds iff +Node_Id identifies
a node of +Tree with -TreeNew being the resultant from removing from
+Tree all the nodes of the subtree rooted at the node with identifier
+Node_Id.

*/

ordTree_removeSubTree(t(Id0,L0,_,_,CHN0),Id,T2) :-
    (   
        Id0==Id 
    -> 
        T2=nil
    ; 
        forest_containingTree(CHN0,Id,Lst1,CH0,Lst2),
        ordTree_removeSubTree(CH0,Id,CH1),
        (    
            CH1==nil 
        -> 
            append(Lst1,Lst2,CHN1)
        ; 
            append(Lst1,[CH1|Lst2],CHN1)
        ),
        ordTree_constructor(Id0,L0,CHN1,T2)
    ).
   
/* 

ordTree_shrinkSubTree(+Tree,+Node_Id,-TreeNew) succeeds iff +Node_Id
identifies a node of +Tree with -TreeNew being the tree resulting from
deleting from +Tree all the nodes of the subtree rooted at the node
with identifier +Node other than the root node.  That is, the node
with identifier +Node_Id is not deleted.

*/

ordTree_shrinkSubTree(t(Id0,L0,_,_,CHN0),Id,T2) :-
    (    
        Id0==Id 
    -> 
        T2=t(Id0,L0,0,1,[])
    ; 
        forest_containingTree(CHN0,Id,Lst1,CH0,Lst2),
        ordTree_shrinkSubTree(CH0,Id,CH1),
        append(Lst1,[CH1|Lst2],CHN1),
        ordTree_constructor(Id0,L0,CHN1,T2)
    ).

ordTree_weight(nil,0).
ordTree_weight(t(_,_,_,W,_),W).

ordTree_height(nil,-1). 
ordTree_height(t(_,_,H,_,_),H).


/* ordTree_leftmost(+Tree,-Node) -Node is the leftmost child of the
root of +Tree */

ordTree_leftmost(t(_,_,_,_,[t(Id,Lb,Ht,Wt)|_]),node(Id,Lb,Ht,Wt)).

/* ordTree_rightmost(+Tree,-Node) -Node is the rightmost child of the
root of +Tree */

ordTree_rightmost(t(_,_,_,_,CHN),N) :-
	forest_rightmost(CHN,CH),
	ordTree_rootNode(CH,N).


/*

ordTree_rootNode(+Tree,-Node) -Node is the root of +Tree.

*/

ordTree_rootNode(t(Id,L,H,W,_),node(Id,L,H,W)). 


/*

ordTree_getNode(+T,-N) -N is a node of +T

*/

ordTree_getNode(t(Id,Lb,Ht,Wt,_),node(Id,Lb,Ht,Wt)).
ordTree_getNode(t(_,_,_,_,ISTS),N) :-
    member(IST,ISTS),
    ordTree_getNode(IST,N). 

ordTree_getLeafNode(t(Id,Lb,0,1,[]),node(Id,Lb,0,1)).
ordTree_getLeafNode(t(_,_,_,_,ISTS),N) :-
    member(IST,ISTS),
    ordTree_getLeafNode(IST,N). 

ordTree_getChildOfRoot(t(_,_,_,_,ISTS),N) :-
    member(IST,ISTS),
    ordTree_rootNode(IST,N).

/*forest_containingTree(+Forest,+Node_Id,-Forest1,-Tree,-Forest2)
+Forest is a list of trees such that each node of the trees in
+Forests is identified with a unique number and a node in a tree in
the front of +Forest is identified with a smaller number than a node
in a tree in the back of +Forest.  +Node_Id is a number.
forest_containingTree(Forest,Node_Id,Forest1,Tree,Forest2) succeeds if
+Node_Id identifies a node in a tree in +Forest with -Forest1
consisting of all the trees in +Forest whose nodes are identified with
numbers smaller than +Node_Id, -Tree being the tree in +Forest
containing the node identified with number +Node_Id, and -Forest2
consisting of all the trees in +Forest whose nodes are identified with
numbers greater than +Node_Id.*/

forest_containingTree([T],_,[],T,[]).
forest_containingTree([T1,t(Id2,L2,H2,W2,CHN2)|F],Id,[],T1,[t(Id2,L2,H2,W2,CHN2)|F]):-
    Id < Id2,
    !.
forest_containingTree([T1,t(Id2,L2,H2,W2,CHN2)|F],Id,[T1|F1],T,F2) :-
%    Id >= Id2,
    forest_containingTree([t(Id2,L2,H2,W2,CHN2)|F],Id,F1,T,F2).


/*forest_height(+Forest,-MaxHeight) MaxHeight is the maximum of the
heights of the trees in Forest.*/

forest_height([],-1).
forest_height([Tree|TreeLst],Max) :-
    ordTree_height(Tree,H),
    forest_height(TreeLst,MaxT),
    (    H>MaxT 
      -> 
         Max is H
       ; 
         Max is MaxT 
    ).

forest_weight([],0).
forest_weight([H|L],WT) :-
    ordTree_weight(H,WH),
    forest_weight(L,WL),
    WT is WL+WH.

forest_rightmost([Rightmost],Rightmost).
forest_rightmost([_,Tree|Forest],Rightmost) :-
	forest_rightmost([Tree|Forest],Rightmost).

