:- use_module(tree23,[empty_23/1]).
:- use_module(testm,	[testm/3, filter/3]).
:- use_module(diagnoser,[diagnoser/3]).
:- use_module(corrector,[corrector/3]).

:- use_module(kb,[clear/0,store/2]). 

:- use_module(aux,[match/3,read_sequence/3]). 

:- use_module(parameters,
              [set_parameter/0,view_parameters/0,read_parameter_value/2]).

:- use_module(store, [
            pg_EMPTY/1,
            pg_SAVE/2,
            pg_LOAD/2,
            pg_SHOW/1,
            pg_ShowPROCS/2,
            pg_ShowPROC/2
          ]).

/* Shell supports the following commands: 

 view                           -> display the values of the parameters

 set                            -> set the values of the parameters

 load                        -> requests the user to provide the
                                   <file_name> and then proceeds as 
                                   command "load <file_name>".

 load <file_name>            -> loads <file_name> as program

 listing                           -> lists the program procedure by 
                                   procedure.

 listing <procedure_name>          -> lists procedure(s) with name 
                                   <procedure_name>

 listing <procedure_name_arity> -> lists procedure with name and arity
                                   <procedure_name_arity>.

 exec                           -> requests the user to provide a 
                                   goal <goal> and then proceeds as 
                                   comman "run <goal>".

 exec <goal>                    -> executes <goal> with the program 

 diagnose                       -> diagnoses the program with the 
                                   symptom just found.

 exit                           -> exits the debugging system.

 symptoms                       -> displays errors detected

*/

/* 
SPECIFICATION
-------------

PUBLIC  go/0 top-level procedure that calls debug/1 with intial
        state that consists of an oracle as specification, an empty
        program, an empty symptom set and a "unknown" bug that indicates
        that no bug has been found so far.

*/ 

go :- 
    pg_EMPTY(EmptyPg),
    help_message,
    debug(state(user,EmptyPg,[])).

/* 
SPECIFICATION
-------------

PRIVATE debug(+/-STATE) reads a command from the user, modifies state
        information and calls itself recursively. A state
        consistes of a specification, a program, a symptom set

*/

debug(State) :- 
    get_command(Command),
    !,
    (  
        Command=quit 
    -> 
        true
    ; 
        exec(Command,State,NewState),
        debug(NewState)
    ).

/* 
SPECIFICATION
-------------

PRIVATE exec(+Command,+State_In,-State_Out) executes Command under
        state OldSate and modifies state from State_In into State_Out.

*/

exec(view,S,S) :-           
    view_parameters.
exec(set,S,S)  :- 
    set_parameter.

exec(symptoms,Si,Si) :-
    !,
    state_getPg(Si,Pg),
    (
        pg_EMPTY(Pg)
    ->
        message('Use "load" to load program.')
    ;
        state_getSS(Si,SS),
        symptoms_SHOW(SS)
    ).

exec(diagnose,Si,So) :-
    !,
    state_getSS(Si,SSi),
    state_getPg(Si,Pi),
    (
        (
            pg_EMPTY(Pi),
            message('Use "load" to load program.')
        ;
            SSi = [],
            message('Use "exec" to detect bug symptoms.')
        )
    ->
        So = Si
    ;
        SSi = [Spt|SS],
        state_getPg(Si,Pi),
        diagnoser(Pi,Spt,Bug),
        corrector(Pi,Bug,Po),
        state_updatePg(Si,Po,S1),
        filter(Po,[Spt|SS],SS1),
        state_updateSS(S1,SS1,So)
    ).

exec(listing,S,S) :-       
    !,     
    state_getPg(S,Pg),
    (
        pg_EMPTY(Pg)
    ->
        message('Use "load" to load program.')
    ;
        pg_SHOW(Pg)
    ).

exec(listing(N/A),S,S) :-
    !,
    state_getPg(S,Pg),
    (
        pg_EMPTY(Pg)
    ->
        message('Use "load" to load program.')
    ;
        pg_ShowPROC(Pg,N/A)
    ).

exec(listing(PN),S,S) :-       
%    \+ PN = _/_,
    !,
    state_getPg(S,Pg),
    (
        pg_EMPTY(Pg)
    ->
        message('Use "load" to load program.')
    ;
        pg_ShowPROCS(Pg,PN)
    ).

exec(load(Term),Si,So) :-  
	!,
	(
		term_to_fn(Term,FN)
	->
		clear,
	        pg_LOAD(FN,Pg),
	        state_updatePg(Si,Pg,S1),
 		state_updateSS(S1,[],So)
	;
		format('~t~w is not a file name.~t~55|~n',Term)
	).

exec(save(Term),S,S) :-  
    	!,
	(
		term_to_fn(Term,FN)
	->
	        state_getPg(S,Pg),
	        (
	 		pg_EMPTY(Pg)
	    	->
        		message('Use "load" to load program.')
    		;
        		pg_SAVE(FN,Pg)
		)
	;
		format('~t~w is not a file name.~t~55|~n',Term)
    	).

exec(exec(Goal),Si,So):- 
    !,
    state_getPg(Si,Pg),
    (
        pg_EMPTY(Pg)
    ->
        message('Use "load" to load program.'),
        So = Si
    ;
        (
           testm(Pg,Goal,SYMPTOM) 
        -> 
           state_getSS(Si,SYMPTOMS),
           state_updateSS(Si,[SYMPTOM|SYMPTOMS],So)
        ; 
           So = Si
        )
    ).


exec(valid(A),Si,Si):- 
    !,
    store(valid,A).

exec(unsatisfiable(A),Si,Si):- 
    !,
    store(unsatisfiable,A).

exec(faithful(A),Si,Si):- 
    !,
    (
        A = PN/PA
    ->
        functor(Call,PN,PA),
        store(faithful,Call)
    ;
        store(faithful,A)
    ).

exec(correct(A),Si,Si):- 
    !,
    (
        A = PN/PA
    ->
        functor(Call,PN,PA),
        store(correct,Call)
    ;
        store(correct,A)
    ).

exec(help,Si,Si) :-
    !,
    help_message.

state_getSp(state(Sp,_Pg,_SS),Sp).
state_getPg(state(_Sp,Pg,_SS),Pg).
state_getSS(state(_Sp,_Pg,SS),SS).

state_updateSp(state(_Sp0,Pg,SS),Sp,state(Sp,Pg,SS)).
state_updatePg(state(Sp,_Pg0,SS),Pg,state(Sp,Pg,SS)).
state_updateSS(state(Sp,Pg,_SS0),SS,state(Sp,Pg,SS)).


symptoms_SHOW([]) :-
    message('Use "exec" to detect bug symptoms.').
symptoms_SHOW([SYMPTOM]) :-
    nl,
    tab(4),
    write('The following is a detected bug symptom.'),
    nl,
    tab(8),write(SYMPTOM), write('.'),nl.
symptoms_SHOW([SYMPTOM1,SYMPTOM2|SYMPTOMS]) :-
    nl, 
    tab(4),
    write('The following are detected bug symptoms.'), 
    nl,
    symptoms_SHOW_1([SYMPTOM1,SYMPTOM2|SYMPTOMS]),
    nl.

symptoms_SHOW_1([]).
symptoms_SHOW_1([SYMPTOM|SYMPTOMS]) :-
    tab(8),write(SYMPTOM),nl,
    symptoms_SHOW_1(SYMPTOMS).

message(Message) :-
   tab(4),write(Message),nl.


help_message :-
    format('~t Declarative Diagnosis ~t ~55|~n',[]),
    format('~t --------------------- ~t ~55|~2n',[]),
    format('~s~20+~`.t~s~55|~n',["correct","assert a procedure is free of bugs"]),
    format('~s~20+~`.t~s~55|~n',["diagnose","bug diagnosis"]),
    format('~s~20+~`.t~s~55|~n',["exec","execute a call"]),
    format('~s~20+~`.t~s~55|~n',["faithful","state a procedure is free of errors"]),
    format('~s~20+~`.t~s~55|~n',["help","display this help menu"]),
    format('~s~20+~`.t~s~55|~n',["list","list program/procedure"]),
    format('~s~20+~`.t~s~55|~n',["load/save","load/save program"]),
    format('~s~20+~`.t~s~55|~n',["quit","quit this system"]),
    format('~s~20+~`.t~s~55|~n',["set/view","set/view system parameters"]),
    format('~s~20+~`.t~s~55|~n',["symptoms","list detected symptoms"]),
    format('~s~20+~`.t~s~55|~n',["valid","assert an atom is valid"]),
    format('~s~20+~`.t~s~55|~n',["unsatisfiable", "assert an atom is unsatisfiable"]).


/*get_command(Command) repeatedly read from screen until a valid command is read.*/

get_command(Cmd) :- 
        cmd_names(CMDS),
        write('debug % '),
        read_sequence([32,46,10],Chars,Last),
        (
	        Chars = []
	->
		get_command(Cmd)
	;
		match(CMDS,Chars,CMDS1),
	        (
        	    	CMDS1 = []
	        ->
 		    	( Last=10 ; skip(10) ),
            		format('~t~s ~s.~t ~s.~t~55|~n',
            			["No command matches",Chars,"Use help to list commands"]),
	            	get_command(Cmd)           
        	;
            		CMDS1 = [CN]
	        ->
        	    	complete_cmd(CN,Last,Cmd)
	        ;   
 		    	( Last=10 ; skip(10) ),
	                length(CMDS1,N),
        	        make_format(N,'~t~s matches ~w commands: ~a',F),
                	format(F,[Chars,N|CMDS1]),
	                get_command(Cmd)           
        	)
	).

cmd_names([
	correct,	diagnose,
	faithful,	help,		listing,	load, 
        quit,	save,   set,    symptoms, 
        exec,  unsatisfiable, 	valid,  view]).

anadic(diagnose).
anadic(help).  anadic(quit). 
anadic(set). anadic(symptoms). 
anadic(view).

cmd_prompt(correct, 		'procedure or goal? '). 
cmd_prompt(faithful,		'procedure or goal? '). 
cmd_prompt(exec, 		'goal? '). 
cmd_prompt(valid,		'atomic goal? ').
cmd_prompt(unsatisfiable,	'atomic goal? ').
cmd_prompt(listing,		'procedure? '). 
cmd_prompt(load,		'file? ').  
cmd_prompt(save,		'file? '). 

make_format(1, F0, F) :-
    concat(F0, '~t~55|~n', F).
make_format(N, F0, F) :-
 	N > 1,
	!,
	concat(F0,', ~a',F1),
        N1 is N-1,
        make_format(N1,F1,F).

%:- use_module(library(charsio),	[chars_to_stream/2]).

complete_cmd(CN,Last,Cmd) :-
    	(
        	anadic(CN) 
        ->
                ( Last = 10 ; skip(10) ), 
		Cmd = CN
        ;
                cmd_prompt(CN,Prompt),
                complete_cmd(CN,Prompt,Last,Cmd)
        ).

complete_cmd(CN,Prompt,Last,Cmd) :-
        (CN=save ; CN=load),
        !,
    	(
            	Last = 32 
        -> 
                read_sequence([32,10],Chars,Last1),
                string_to_list(Arg,Chars),
                (
                        Chars \= [],Chars \= [46]
                ->
                        Cmd =.. [CN,Arg] 
                ; 
                       ( Last1 = 10 ; skip(10) ),
                        write(Prompt),
                        complete_cmd(CN,Prompt,32,Cmd)
                )
	;
                ( Last = 10 ; skip(10) ),
		(
			CN = listing
		->
			Cmd = CN
		;
	                write(Prompt),
        	        complete_cmd(CN,Prompt,32,Cmd)
                )
	).

complete_cmd(CN,Prompt,Last,Cmd) :-
    	(
            	Last = 32 
        -> 
                (
	                read_term(Arg, [syntax_errors(quiet)])
                ->
                        Cmd =.. [CN,Arg] 
                ;       
                        write(Prompt),
                        complete_cmd(CN,Prompt,32,Cmd)
                )
	;
                ( Last = 10 ; skip(10) ),
		(
			CN = listing
		->
			Cmd = CN
		;
	                write(Prompt),
        	        complete_cmd(CN,Prompt,32,Cmd)
                )
	).

term_to_fn(Term,FN) :-
    ground(Term),
    !,
    term_to_str(Term,Str),
    name(FN,Str).

term_to_str(Prefix/Atom,Str) :-
	!,
	atomic(Atom),
    	term_to_str(Prefix,Pf),
    	append(Pf,"/",Pf1),
	(
		name(Atom,Str1),
		append(Pf1,Str1,Str)
	).
term_to_str(Atom,Str) :-
	atomic(Atom),
    	name(Atom,Str).


