:- module(strategies,
          [non_root_node/3,
           branch_node/3
          ]
         ).

:- use_module(goal,[goal_select/3]).

:- use_module(orderedTree,[
               ordTree_getLeafNode/2, 
               node_isLeaf/1,
               ordTree_getChildOfRoot/2,
               node_isSolutionNode/1,
               forest_weight/2,
               ordTree_height/2
              ]).



/*
SPECIFICATION
------------- 

This  module  contains   procedures  that  implement  various  strategies  for
searching I-congruent partial proof trees and  I-complete partial SLDNF trees.
The name  for  each  procedure has three  parts.  The  first part  of the name
indicates the kind of the trees that  the named procedure searches.   If it is
"cps"  then  the corresponding procedure  searches  I-complete  partial  proof
trees. And if it is "cpp" then the corresponding procedure searhes I-congruent
proof trees.  The second part of the name is "search" for every procedure. The
third  part  of  the  name indicates  the  strategies  that  the corresponding
procedure  uses  to  search.   "dq"  means  the  "divide-and-query"   strategy
developed  by (Shapiro 82). "td" means the  "top-down"  strategy  described in
(Lloyd 87) and "bu" means the "bottom-up" strategy also found in (Lloyd 87).

Each I-congruent  partial  proof tree  search  procedure  is applicable to any
I-congruent partial proof tree with  height greater than 1 and each I-complete
partial SLDNF tree search procedure is  applicable to  any I-complete  partial
SLDNF  tree with height  greater than 0. This requirement does not  compromise
the  completeness  of  the  diagnosing  procedures  because  these  diagnosing
procedures  checks  the  height of  the corresponding  trees before  they call
searching procedures and deal with the cases that this requirement is not met.

PUBLIC cps_search_dq(+Tree,-Node) 
       cpp_search_dq(+Tree,-Node)
      
PUBLIC cps_search_td(+Tree,-Node)
       cpp_search_td(+Tree,-Node)

PUBLIC cps_search_bu(+Tree,-Node)
       cpp_search_bu(+Tree,-Node)

PUBLIC cps_search_tdz(+Tree,-Node)
       cpp_search_tdz(+Tree,-Node)


*/ 

non_root_node(top_down,T,N) :- 
    cps_search_td(T,N).
non_root_node(bottom_up,T,N) :-
    cps_search_bu(T,N).
non_root_node(divide_and_conquer,T,N) :-
    cps_search_dq(T,N).
non_root_node(top_down_zooming,T,N) :-
    cps_search_tdz(T,N).

branch_node(top_down,T,N) :-
    cpp_search_td(T,N).
branch_node(bottom_up,T,N) :-
    cpp_search_bu(T,N).
branch_node(divide_and_conquer,T,N) :-
    cpp_search_dq(T,N).
branch_node(top_down_zooming,T,N) :-
    cpp_search_tdz(T,N).

/* 

The  following procedures  implement divide-and-query strategies for searching
cpp and cps trees. Each time  a cpp tree  or  a cps  tree  is searched, a node
which roughly divides the tree into two roughly equivalent parts is selected.

*/

cps_search_dq(t(Id,Lb,Ht,Wt,CHN),Nm) :-
    Ht =\= 0,
    Wtm is Wt // 2,
    ordTree_getChildOfRoot(t(Id,Lb,Ht,Wt,CHN),Nm0),
    \+ node_isSolutionNode(Nm0),
    cps_search_dq_1_list(Wtm,Nm0,CHN,Nm).

cps_search_dq_1(Wtm,node(Id0,Lb0,Ht0,Wt0),t(Id,Lb,Ht,Wt,CHN),Nm) :-
    D  is (Wtm - Wt)*(Wtm - Wt),
    D0 is (Wtm - Wt0)*(Wtm - Wt0),
    (    
        (Wt<Wt0,D>=D0) 
    ->  
        Nm=node(Id0,Lb0,Ht0,Wt0) % not to search too small a tree 
    ; 
        (    
            D<D0,\+ node_isSolutionNode(node(Id,Lb,Ht,Wt))
        -> 
            Nm1=node(Id,Lb,Ht,Wt)
        ;
            Nm1=node(Id0,Lb0,Ht0,Wt0)
        ),
        cps_search_dq_1_list(Wtm,Nm1,CHN,Nm)
    ).

cps_search_dq_1_list(_,Nm0,[],Nm0). 
cps_search_dq_1_list(Wtm,Nm0,[CH|CHN],Nm) :-
    cps_search_dq_1(Wtm,Nm0,CH,Nm1),
    cps_search_dq_1_list(Wtm,Nm1,CHN,Nm).

cpp_search_dq(t(Id,Lb,Ht,Wt,CHN),Nm) :-
    Ht > 1,
    (
        ordTree_getChildOfRoot(t(Id,Lb,Ht,Wt,CHN),Nm0),
        \+ node_isLeaf(Nm0)
    ->
        Wtm is Wt // 2,
        cpp_search_dq_1_list(Wtm,Nm0,CHN,Nm)
    ).

cpp_search_dq_1(Wtm,node(Id0,Lb0,Ht0,Wt0),t(Id,Lb,Ht,Wt,CHN),Nm) :-
    D is (Wtm - Wt)*(Wtm - Wt),
    D0 is (Wtm - Wt0)*(Wtm - Wt0),
    (    
        ( Wt<Wt0,D>=D0 )
    ->
        Nm=node(Id0,Lb0,Ht0,Wt0)% not to search too small a tree 
    ; 
        (    
            D<D0,\+ node_isLeaf(node(Id,Lb,Ht,Wt))
        -> 
            Nm1=node(Id,Lb,Ht,Wt)
        ; 
            Nm1=node(Id0,Lb0,Ht0,Wt0)
        ),
        cpp_search_dq_1_list(Wtm,Nm1,CHN,Nm)
    ).

cpp_search_dq_1_list(_,Nm0,[],Nm0). 
cpp_search_dq_1_list(Wtm,Nm0,[CH|CHN],Nm) :-
    cpp_search_dq_1(Wtm,Nm0,CH,Nm1),
    cpp_search_dq_1_list(Wtm,Nm1,CHN,Nm).

/* 

The following procedures implement top-down strategies for searching
cpp or cps trees. Each time a cpp or cps tree is searched, a child of
the root of the tree is selected.

*/

cps_search_td(T,N) :-
    \+ ordTree_height(T,0),
    ordTree_getChildOfRoot(T,N),
    \+ node_isSolutionNode(N).
cpp_search_td(T,N) :-
    ordTree_height(T,Ht),
    Ht>1,
    ordTree_getChildOfRoot(T,N),
    \+ node_isLeaf(N).

/* 

The following procedures implement strategies for searching cpp and cps trees.
Each time a  cpp  tree  is searched,  a node that  is the parent  of a leaf is
selected. Each time a cps tree is searched, a leaf that is not a solution node
is  selected. If the tree does  not have such a leaf then a node that  is the 
parent of a leaf is selected instead.

*/

cps_search_bu(T,N) :- 
    \+ ordTree_height(T,0),
    (    (ordTree_getLeafNode(T,N),\+ node_isSolutionNode(N)) 
    ->   
         true
    ; 
         cpp_search_bu(T,N)   
    ).

cpp_search_bu(t(_,_,Ht,_,CHN),Node) :-
    Ht > 1,    
    cpp_search_bu_list(CHN,Node).

%cpp_search_bu_list([t(Id0,Lb0,Ht0,Wt0,CHN0)|CHN],Node) :-
%        %non-deteministic version
%    (    
%        cpp_search_bu_list(CHN0,Node)
%    ;
%        Ht0 == 1,
%        Node = node(Id0,Lb0,Ht0,Wt0)
%    ; 
%        cpp_search_bu_list(CHN,Node)
%    ).

cpp_search_bu_list([t(Id0,Lb0,Ht0,Wt0,CHN0)|CHN],Node) :-
    (    
        Ht0 == 0
    -> 
        cpp_search_bu_list(CHN,Node)
    ;
        Ht0 == 1
    -> 
        Node = node(Id0,Lb0,Ht0,Wt0)
    ; 
        cpp_search_bu_list(CHN0,Node)
    ).

/* 

The following procedures  implement  top-down zooming strategies.  Each time a
cpp tree  is searched, a node whose literal has  the same predicate descriptor
(name and arity) as the literal of the root is selected if the tree has such a
node.  Otherwise, a  child of  the root that is not a  leaf is  selected. Each
time a  cps  tree is searched, a  node whose  selected  literal  has the  same
predicate descriptor  as the  selected literal of the root  is  slected if the
tree  has such a  node. Otherwise, a child of the  root that is not a solution
node is selected.

*/

cpp_search_tdz(t(Id,Lb,Ht,Wt,CHN),Node) :-
    Ht>1,
    predicate_descriptor(Lb,PD),
    (     
        cpp_zoom_list(CHN,PD,Node)
    ->
        true
    ;
        ordTree_getChildOfRoot(t(Id,Lb,Ht,Wt,CHN),Node),
        \+ node_isLeaf(Node)
    ).
 
cpp_zoom(t(Id,Lb,Ht,Wt,CHN),PD,Node) :-
    Ht =\= 0,
    (
        predicate_descriptor(Lb,PD)
    ->  
        Node = node(Id,Lb,Ht,Wt)
     ;   
        cpp_zoom_list(CHN,PD,Node)
    ).

cpp_zoom_list([T|TL],PD,Node) :-
    (
        cpp_zoom(T,PD,Node) 
    ->  
        true
    ;  
        cpp_zoom_list(TL,PD,Node)
    ).

%cpp_search_tdz(t(Id,Lb,Ht,Wt,CHN),Node) :-
%        %non-deterministic version
%    predicate_descriptor(Lb,PD),
%    (     
%        cpp_zoom_list(CHN,PD,Node)
%    ;
%        ordTree_getChildOfRoot(t(Id,Lb,Ht,Wt,CHN),Node),
%        \+ node_isLeaf(Node)
%    ).
% 
%cpp_zoom(t(Id,Lb,Ht,Wt,CHN),PD,Node) :-
%    \+ ordTree_height(t(Id,Lb,Ht,Wt,CHN),0),
%    (
%        predicate_descriptor(Lb,PD),
%        Node = node(Id,Lb,Ht,Wt)
%     ;   
%        cpp_zoom_list(CHN,PD,Node)
%    ).
%
%cpp_zoom_list([T|TL],PD,Node) :-
%    (
%        cpp_zoom(T,PD,Node) 
%    ;  
%        cpp_zoom_list(TL,PD,Node)
%    ).

cps_search_tdz(t(Id,W/Sub,Ht,Wt,CHN),Node) :-
    Ht =\= 0,
    goal_select(W,L,_),
    predicate_descriptor(L,PD),
    (     
        cps_zoom_list(CHN,PD,Node)
    ->
        true
    ;
        ordTree_getChildOfRoot(t(Id,W/Sub,Ht,Wt,CHN),Node),
        \+ node_isSolutionNode(Node)
        % remark: no solution node in a cps for insufficiency symptom
    ).
 
cps_zoom(t(Id,W/Sub,Ht,Wt,CHN),PD,Node) :-
    goal_select(W,L,_),    
    (
        predicate_descriptor(L,PD)
    ->  
        Node = node(Id,W/Sub,Ht,Wt)
    ;   
        cps_zoom_list(CHN,PD,Node)
    ).

cps_zoom_list([T|TL],PD,Node) :-
    (
        cps_zoom(T,PD,Node) 
    ->  
        true
     ;  
        cps_zoom_list(TL,PD,Node)
    ).

%cps_search_tdz(t(Id,W/Sub,Ht,Wt,CHN),Node) :-
        % non-deterministic version
%    goal_select(W,L,_),
%    predicate_descriptor(L,PD),
%    (     
%        cps_zoom_list(CHN,PD,Node)
%    ;
%        ordTree_getChildOfRoot(t(Id,W/Sub,Ht,Wt,CHN),Node),
%        \+ node_isSolutionNode(Node)
%    ).
% 
%cps_zoom(t(Id,W/Sub,Ht,Wt,CHN),PD,Node) :-
%    goal_select(W,L,_),    
%    (
%        predicate_descriptor(L,PD),
%        Node = node(Id,W/Sub,Ht,Wt)
%    ;   
%        cps_zoom_list(CHN,PD,Node)
%    ).
%
%cps_zoom_list([T|TL],PD,Node) :-
%    (
%        cps_zoom(T,PD,Node) 
%     ;  
%        cps_zoom_list(TL,PD,Node)
%    ).

/*predicate_descriptor(+L,?(PN/PA)) +L is  a literal. ?PN  is the name
of the predicate in +L and ?PA is its arity.*/

predicate_descriptor(L,PN/PA) :-
    (
        L = (\+A) 
    -> 
        functor(A,PN,PA)
    ;
        functor(L,PN,PA)
    ).


/*

cpp :-
    P = [(qs([X|L],L0) :-
    part(L,X,L1,L2),
    qs(L1,L3),
    qs(L2,L4),
    ap([X|L3],L4,L0)         %(1) ap(L3,[X|L4],L0)
    ), 
    (qs([],[]):-true),
    (part([X|L],Y,L1,[X|L2]):-
                             %(2)  Y<X, missing
    part(L,Y,L1,L2)),
    (part([X|L],Y,[X|L1], L2) :-
    Y=<X,                    %(3)  Y >= X,
    part(L,Y,L1,L2)),
    (part([],X,[],[]):-true),
    (ap([X|L1],L2,[X|L3]) :-
    ap(L1,L2,L3)),
    (ap([],L,L):-true)],
    icppt(P,qs([2,1],[2,1]),Tree),
    ordTree_display(Tree,0),nl,nl,
    branch_node(Tree,N),
    write(N),nl,fail.

cps :-
    P = [
            (qs([X|L],L0) :-
                part(L,X,L1,L2),
                qs(L1,L3),
                qs(L2,L4),
                ap([X|L3],L4,L0)          %(1) ap(L3,[X|L4],L0)
            ),
            
            (qs([],[]) :-
                true
            ),
            
            (part([X|L],Y,L1,[X|L2]) :- 
                                            % Y<X,
                part(L,Y,L1,L2)
            ),
            (part([X|L],Y,[X|L1], L2) :- 
                Y=<X,                       % Y>=X,
                part(L,Y,L1,L2)
            ),
            (part([],X,[],[]):-
                true
            ),
            (ap([X|L1],L2,[X|L3]) :-
                ap(L1,L2,L2)                % ap(L1,L2,L3) 
            ),
            (ap([],L,[]):-                  % ap([],L,[])
                true
            )
        ],
    icpst(P,qs([2,1],_Sorted),Tree),
    ordTree_display(Tree,0),nl,nl,
    non_root_node(Tree,N),
    write(N),nl,fail.
*/
