:- module(substitution, [ 
	mgu/3,
	sub_compose/3,
        sub_constructor/3,
        sub_domain/2,
        sub_gen_renaming/2,
        sub_genSymbTb/3,
        sub_skolemize/3,
	sub_apply/3,
        sub_inverse/2,
        sub_display/1,
	sub_restrict/3,
        subs_display/1,
	subs_restrict/3,
        variant/2
   ]).

:- use_module(aux,[vars/2]).

:- use_module(ordsets, 	 	[
	ord_add_element/3,
	ord_subset/2
   ]).


mgu(X,Y,[]) :- 
    X==Y,
    !.
mgu(X,Y,[X/Y]) :-
    var(X),
    !,
    \+ occur(X,Y).
mgu(X,Y,[Y/X]) :-
    var(Y),
    !,
    \+ occur(Y,X).
mgu(X,Y,Substs) :-
    !,
    X=..[F|Args1],
    Y=..[F|Args2],
    mgu_list(Args1,Args2,Substs).

mgu_list([],[],[]). 
mgu_list([H1|T1],[H2|T2],Substs) :-
    mgu_list(T1,T2,SubT),
    sub_apply(SubT,H1,H11),
    sub_apply(SubT,H2,H21),
    mgu(H11,H21,SubH),
    sub_compose(SubT,SubH,Substs). 

% occur(+V,+T) +V occurs in term +T.

occur(X,Y) :-	% Y is a variable 
    var(Y),
    !,
    X==Y.
occur(X,Y) :-	% Y is not a variable
    Y=..[_|Args],
    occur_in_one_term(X,Args).

occur_in_one_term(X,[H|_])  :- 
    occur(X,H),
    !.
occur_in_one_term(X,[_|T])  :-
    occur_in_one_term(X,T).

/*sub_domain(+Sub,-VList)  VList is  the  domain  of  Sub  in  the form of
ordered list*/

sub_domain([],[]).
sub_domain([V/_|Sub1],D) :-
    sub_domain(Sub1,D1),
    ord_add_element(D1,V,D).

/*mapsto(+Sub,+Var,-Val)  If  Var  is   in  the  domain  of  Sub  then
mapsto(Sub,Var,Val) succeeds and Val is  the value assigned to Var  by
Sub. Otherwise mapsto(Sub,Var,Val) fails.*/

mapsto([X/V|_],Var,Val) :-
    Var == X,
    !,
    Val = V.
mapsto([_|Sub],Var,Val) :-
    !,
    mapsto(Sub,Var,Val).

/*sub_apply(+Sub,+Ti,-To) To=Sub(Ti).*/

sub_apply(Sub,Ti,To) :-
    (    var(Ti) 
      ->
         ( mapsto(Sub,Ti,To) -> true; To=Ti )
       ; 
         Ti =.. [F|Argsi],
         sub_apply_args(Sub,Argsi,Argso),
         To =.. [F|Argso]         
    ).

sub_apply_args(_,[],[]) :-
    !.
sub_apply_args(Sub,[Argi|Argsi],[Argo|Argso]) :-
    sub_apply(Sub,Argi,Argo),
    sub_apply_args(Sub,Argsi,Argso).

/*sub_restrict(+Subi, +VL,-Subo) Subo = Subi | VL */

sub_restrict(Subi,VL,Subo) :-
    sort(VL,VL1),
    sub_restrict_1(Subi,VL1,Subo).

sub_restrict_1([],_,[]).
sub_restrict_1([V/T|Subi], VL, [V/T|Subo]) :-
    ord_subset([V],VL),
    !,
    sub_restrict_1(Subi,VL,Subo).
sub_restrict_1([_|Subi],VL,Subo) :-
%    \+ ord_subset([V],VL),
    sub_restrict_1(Subi,VL,Subo).

/*subs_restrict(+SubLstIn,+VL,-SubLstOut)*/

subs_restrict(SubLstIn,VL,SubLstOut) :-
    sort(VL,VL1),
    subs_restrict_1(SubLstIn,VL1,SubLstOut).

subs_restrict_1([],_,[]).
subs_restrict_1([Si|SiL],VL,[So|SoL]) :-
    sub_restrict_1(Si,VL,So),
    subs_restrict_1(SiL,VL,SoL).

/* sub_compose(+Theta,+Sigma,-Tau) Tau = Theta o Sigma */

sub_compose(Theta,Sigma,Tau) :-
    instantiation_pass(Theta,Sigma,Tau_),
    merging_pass(Sigma,Tau_,Tau).

instantiation_pass([],_,[]) :-
    !.
instantiation_pass([V/Ti|Theta],Sigma,[V/To|Tau]) :-
    sub_apply(Sigma,Ti,To),
    instantiation_pass(Theta,Sigma,Tau).

merging_pass(Sigma,Theta,Tau) :-
    sub_domain(Theta,VL),
    merging_pass_1(Sigma,VL,Theta,Tau).

merging_pass_1([],_,Theta,Theta).
merging_pass_1([V/T|Sigma],VL,Theta,Tau) :-
    ( 
        ord_subset([V],VL)
    ->
        merging_pass_1(Sigma,VL,Theta,Tau)
    ;
        merging_pass_1(Sigma,VL,Theta,Tau1),
        Tau = [V/T|Tau1]
    ).

/*sub_constructor(+Variable_Name_List,+Term_List,-Substitution)*/

sub_constructor(Vars,Terms,Sub) :-
    	sub_constructor_1(Vars,Terms,Sub1),
	sort(Sub1,Sub).

sub_constructor_1([],[],[]).
sub_constructor_1([VN|VNL],[TM|TML],[VN/TM|Sub]) :-
    	sub_constructor_1(VNL,TML,Sub).


/*subs_display(+Substitution_List) +Substitution_List is a list
of substitutions.  subs_display(Substitution_List) displays the
substitutions in +Substitution_List line by line.*/

subs_display([]).
subs_display([Sub|Sub_List]) :-
    sub_display(Sub),
    subs_display(Sub_List).

/*PRIVATE    sub_display(+Substitution)    +Substituion    is     a
substitution.   sub_display(Substitution)  displays  in   pair  the
variable and the term of each variable assignment in +Substitution.*/

sub_display([]) :-
    nl.
sub_display([Var/Val|Sub]) :-
    tab(4),write(Var),write(' = '),write(Val), 
    sub_display(Sub).

/*sub_gen_renaming(+VL,-Sub) VL is a list of variables,  -Sub is a renaming
substitution with dom(Sub) = VL.*/

sub_gen_renaming([],[]).  
sub_gen_renaming([V|VL],[V/_V1|S]) :-
    sub_gen_renaming(VL,S).  


/*sub_inverse(+Subi,-Subo) Subo is the inverse of Subi*/

sub_inverse([],[]).
sub_inverse([X/T|Subi],Subo) :-
    sub_inverse(Subi,Subo1),
    ord_add_element(Subo1,T/X,Subo).   
    
/* skolemize(+Term,+Reserved,-Skolemized) */

sub_skolemize(Term,Reserved,Skolemized) :-
	sub_genSymbTb(Term,Reserved,SymbolTable),
        sub_apply(SymbolTable,Term,Skolemized).

/*sub_genSymbTb(+Term,-SymbolTable)  generates a mnemoric name for each
variable in +Term. -SymbolTable is a list. Each element is of the form
Var/Name  where Var  is  a variable  in +Term  and  Name  is  the
mnemoric name of Var.*/

sub_genSymbTb(Term,Reserverd, SymbolTable) :-
      vars(Term,Vars),
      mnemoric(Vars,Reserverd,SymbolTable).

/*PRIVATE  mnemoric(+Vars,-Substitution) generates a  mnemoric name
for a variable  in +Vars.  The variable-mnemoric pairs constitute a
-Substitution.*/

mnemoric(Vars,Reserverd,Substitution) :-
    	mnemoric(Vars,Reserverd,Substitution,23). 
mnemoric([],_,[],_).        
mnemoric([V|VL],Reserverd,[V/VN|SUB],Num0) :- 
	gen_vn(Num0,Reserverd,Num,VN),
    	mnemoric(VL,[VN|Reserverd],SUB,Num).

renaming([]).
renaming([_/T|Sub]) :-
    var(T),
    renaming(Sub).

variant(A,B) :-
    mgu(A,B,S),
    renaming(S).

int_to_vn(Int,VN) :-
    Letter is 65 + Int mod 26,
    Int1 is Int // 26,
    digits(Int1, Digits),
    name(VN,[Letter|Digits]).
digits(0,[]).
digits(N,[D|Ds]) :-
    N > 0,
    D is 48 + N mod 10,
    N1 is N // 10,
    digits(N1,Ds).

gen_vn(N0,Reserverd,N,VN) :-
    	int_to_vn(N0,VN1),
    	N1 is N0 + 1,
    	(
		member(VN1,Reserverd)
	->
		gen_vn(N1,Reserverd,N,VN)	
	;
		N = N1, VN = VN1
    	).


