:- module(testm, [testm/3,filter/3]).

:- use_module(oracle,[
       valid/2,
       query_exists/2]).

:- use_module(kb,[valid_kb/1]).

:- use_module(meta,[solve_db/3]).

:- use_module(aux,[term_rename/2]).

/* 
testm(+Program,+Goal,-Symptom)  +Program   is  tested  with   +Goal.
testm(Program,Goal,Symptom)  returns with each -Symptom  that can be
found.
*/

testm(P,A,S) :-
    (  %Now try to find an insufficiency symptom using the partial intended
       %semantics only. The method is to check if an instance of the
       %goal that has been known valid can be proved by the program
       %without instantiating any variable of it. If such an
       %instance of the goal can not be proved without instantiating
       %any variable of it then an insufficiency symptom has been
       %found. If a stack overflow has occured when the instance of
       %the goal is proved then a stack overflow symptom has been
       %found. This part of the testing continues until either a bug
       %symptom has been found or all the instances of the goal that
       %are known valid have not been checked.

       valid_kb(A), 
       numbervars(A,0,_),       % (\forall A) must be provable
       (   
          solve_db(P,A,R)
       -> 
          (  R=err(Cat) -> S=operational(Cat,A) ) 
       ;  
          S = declarative(miss,A)
       )
    ;  %Now try to find a wrong symptom using the partial intended
       %semantic and the oracle. The method is to repeatedly find a
       %computed instance of the goal and check if this computed
       %instance is valid or not. If it is not valid then it is an
       %inconsistency symptom. Should a stack overflow occur, a
       %stack overflow symptom has been found.  This process
       %continues until either no more computed instance of the goal
       %can be found or a bug symptom has been found.
 
       term_rename(A,A0),   %avoid instantiating operational symptom
       solve_db(P,A0,R),      %            
       (                               %
            R = err(Cat)         %
       ->                              %
            S = operational(Cat,A)    %not instantiated
       ;
          (  
             \+ valid(A0,P) 
          -> 
             S = declarative(wrong,A0)
          )
       )
    ;  %Now try to find an insufficiency syptom with help of the oracle.
       %The method is to display all the instances of the goal that
       %are known valid and ask the oracle to check if they are
       %complete.  If they are complete then no insufficency symptom
       %can be found. Otherwise the oracle is requested to provide a
       %valid anwser of the goal that are not already known and a
       %insufficiency symptom has been found.

       query_exists(A,true),
       S = declarative(miss,A)
    ),
    new_symptom_SHOW(S).

/*filter(+Program,+Symptoms_In,-Symptoms_Out) +Symptoms_In is a set of
symptoms of a previous version of +Program.  -Symptoms_Out is a subset
of +Symptoms_In and -Symptoms_Out is a set of symptoms of +Program.*/

filter(_Program,[],[]).
filter(Program,[declarative(wrong,A)|SYMPTOMS_In],SYMPTOMS_OUT) :-
    (  
        \+ (numbervars(A,0,_),solve_db(Program,A,true)) 
    -> 
        SYMPTOMS_OUT = SYMPTOMS_OUT1
    ; 
        old_symptom_SHOW(declarative(wrong,A)),
        SYMPTOMS_OUT = [declarative(wrong,A)|SYMPTOMS_OUT1] 
    ),
    filter(Program,SYMPTOMS_In,SYMPTOMS_OUT1).
        
filter(Program,[declarative(miss,A)|SYMPTOMS_In],SYMPTOMS_OUT) :-
    (  
        \+ (numbervars(A,0,_),solve_db(Program,A,true)) 
    -> 
        old_symptom_SHOW(declarative(miss,A)),
        SYMPTOMS_OUT = [declarative(miss,A)|SYMPTOMS_OUT1]
    ;  
        SYMPTOMS_OUT = SYMPTOMS_OUT1 
    ),
    filter(Program,SYMPTOMS_In,SYMPTOMS_OUT1). 

filter(Program,[operational(Cat,A)|SYMPTOMS_In],SYMPTOMS_OUT) :-
    (  
        \+ solve_db(Program,A,err(Cat))
    -> 
        SYMPTOMS_OUT = SYMPTOMS_OUT1
    ; 
        old_symptom_SHOW(operational(Cat,A)),
        SYMPTOMS_OUT = [operational(Cat,A)|SYMPTOMS_OUT1]
    ),
    filter(Program,SYMPTOMS_In,SYMPTOMS_OUT1).

old_symptom_SHOW(Symptom) :-
    tab(4),
    write(Symptom),
    write(' has not been removed.'),
    nl.
new_symptom_SHOW(Symptom) :-
    tab(4),
    write(Symptom),
    write(' has been detected.'),
    nl.
    